/*
 *  linux/drivers/video/pm3fb.c -- 3DLabs Permedia3 frame buffer device
 *  
 *  Copyright (C) 2001 Romain Dolbeau <dolbeau@irisa.fr>
 *  Based on code written by:
 *           Sven Luther, <luther@dpt-info.u-strasbg.fr>
 *           Thomas Witzel, <twitzel@nmr.mgh.harvard.edu>
 *           Alan Hourihane, <alanh@fairlite.demon.co.uk>
 *  Based on linux/drivers/video/skeletonfb.c:
 *	Copyright (C) 1997 Geert Uytterhoeven
 *  Based on linux/driver/video/pm2fb.c:
 *  Copyright (C) 1998-1999 Ilario Nardinocchi (nardinoc@CS.UniBO.IT)
 *  Copyright (C) 1999 Jakub Jelinek (jakub@redhat.com)
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 *
 *  CHANGELOG:
 *  Wed Feb 21 19:58:56 CET 2001, v 1.0.1 : OpenFirmware support, fixed memory detection, better debug support, code cleaning
 *  Wed Feb 21 14:47:06 CET 2001, v 1.0.0 : First working version
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/fb.h>
#include <linux/init.h>

#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>

#include <asm/io.h>
#include <linux/pci.h>
#ifdef CONFIG_FB_OF
#include <asm/prom.h>
#endif
#include "pm3fb.h"

/* in case it's not in linux/pci.h */
#ifndef PCI_DEVICE_ID_3DLABS_PERMEDIA3
#define PCI_DEVICE_ID_3DLABS_PERMEDIA3 0x000a
#endif

/* not sure if/why it's needed. doesn't work without on my PowerMac... */
#ifdef __BIG_ENDIAN
#define MUST_BYTESWAP
#endif

#define PM3_MAX_BOARD 4

/* for driver debugging ONLY */
/* 0 = assert only, 1 = error, 2 = info, 3+ = verbose */
#define PM3FB_MASTER_DEBUG 2
#if defined(PM3FB_MASTER_DEBUG) && (PM3FB_MASTER_DEBUG >= 3)
#define PM3FB_TRACE
#endif


#ifdef PM3FB_MASTER_DEBUG
#define DPRINTK(l,a,b...) do { if ((l) <= PM3FB_MASTER_DEBUG) printk("pm3fb: %s: " a, __FUNCTION__ , ## b); } while (0)
#define DASSERT(t,a,b...) do { if (!(t)) printk("pm3fb: _assert failed: %s: " a, __FUNCTION__ , ## b); } while (0)
#ifdef PM3FB_TRACE
#define DTRACE printk("pm3fb: _enter %s\n", __FUNCTION__)
#else /* PM3FB_TRACE */
#define DTRACE
#endif /* PM3FB_TRACE */
#else
#define DPRINTK(l,a,b...)
#define DASSERT(t,a,b...)
#define DTRACE
#endif 

/* ************************************* */
/* ***** The various "global" data ***** */
/* ************************************* */

/* those will need a rework for multiple board support */
/* Driver name */
static const char permedia3_name[16]="Permedia3";

/* the fb_info struct, mandatory */
struct pm3fb_info
{
  struct fb_info_gen gen;
  unsigned long board_num;
  struct pci_dev *dev;
#ifdef CONFIG_FB_OF
  struct device_node *dn;
#endif
  unsigned char* fb_base;	/* framebuffer memory base */
  u32 fb_size;		/* framebuffer memory size */
  unsigned char* p_fb;	/* physical address of frame buffer */
  unsigned char* v_fb;	/* virtual address of frame buffer */
  unsigned char *pIOBase;/* physical address of registers region, must be rg_base or rg_base+PM2_REGS_SIZE depending on the host endianness */
  unsigned char *vIOBase; /* address of registers afer ioremap() */
  unsigned long pm3_regs[0x2000];
  unsigned long dac_regs[0x100];
  /*  unsigned char cmap[0x300]; */
  unsigned char BppShift;
  struct {
	u8 transp;
	u8 red;
	u8 green;
	u8 blue;
  } palette[256];
	union {
#ifdef FBCON_HAS_CFB16
		u16 cmap16[16];
#endif
#ifdef FBCON_HAS_CFB32
		u32 cmap32[16];
#endif
	} cmap;
};

/* the fb_par struct, mandatory */
/* straight from pm2fb */
struct pm3fb_par {
	u32 pixclock;		/* pixclock in KHz */
	u32 width;		/* width of virtual screen */
	u32 height;		/* height of virtual screen */
	u32 hsstart;		/* horiz. sync start */
	u32 hsend;		/* horiz. sync end */
	u32 hbend;		/* horiz. blank end (also gate end) */
	u32 htotal;		/* total width (w/ sync & blank) */
	u32 vsstart;		/* vert. sync start */
	u32 vsend;		/* vert. sync end */
	u32 vbend;		/* vert. blank end */
	u32 vtotal;		/* total height (w/ sync & blank) */
	u32 stride;		/* screen stride */
	u32 base;		/* screen base (xoffset+yoffset) */
	u32 depth;		/* screen depth (8, 16, 24 or 32) */
	u32 video;		/* video control (hsync,vsync) */
};

/* regular resolution database*/
static struct
{
  char name[16];
  struct pm3fb_par user_mode;
} mode_base[] __initdata =
{
  {"800x600-75", {49500,800,600,16,96,256,1056,1,4,25,625,800,0,8,
		  PM3VideoControl_ENABLE|PM3VideoControl_HSYNC_ACTIVE_HIGH|PM3VideoControl_VSYNC_ACTIVE_HIGH|PM3VideoControl_PIXELSIZE_8BIT}},
  {"1024x768-74", {77825,1024,768,16,176,276,1300,4,9,41,809,1024,0,8,
		   PM3VideoControl_ENABLE|PM3VideoControl_HSYNC_FORCE_HIGH|PM3VideoControl_VSYNC_FORCE_HIGH|PM3VideoControl_PIXELSIZE_8BIT}},
  {"\0", }
};

/* more mandatory stuff (see skeletonfb.c + framebuffer driver HOWTO */
static unsigned long current_mode __initdata = 0;

static struct pm3fb_info fb_info[PM3_MAX_BOARD];
static unsigned long detected_board = 0;
static struct pm3fb_par current_par;
static int current_par_valid = 0;
static struct display disp;

/* really, inside display */
/* static struct fb_var_screeninfo default_var; */

static int currcon = 0;
static int inverse = 0;


/* ******************************************** */
/* ***** A bunch of register-access macro ***** */
/* ******************************************** */
#define PM3_WAIT(n) \
do{ \
	while(PM3_READ_REG(PM3InFIFOSpace)<(n)); \
} while(0)

#define PM3_SLOW_WRITE_REG(r,v)	\
do{                             \
    DASSERT((l_fb_info->vIOBase != (unsigned char*)(-1)), "l_fb_info->vIOBase mapped in slow write\n"); \
	mb();                   \
	PM3_WAIT(1);            \
	mb();                   \
    PM3_WRITE_REG(r,v);     \
} while(0)

#define PM3_SET_INDEX(index) \
do{ \
	PM3_SLOW_WRITE_REG(PM3RD_IndexHigh,(((index)>>8)&0xff)); \
	PM3_SLOW_WRITE_REG(PM3RD_IndexLow,((index)&0xff)); \
} while(0)

#ifdef MUST_BYTESWAP
#define PM3_READ_REG(r) __swab32(readl((l_fb_info->vIOBase + r)))
#define PM3_WRITE_REG(r, v) writel(__swab32(v), (l_fb_info->vIOBase + r))
#else /* MUST NOT BYTESWAP */
#define PM3_WRITE_REG(r, v) writel(v, (l_fb_info->vIOBase + r))
#define PM3_READ_REG(r) readl((l_fb_info->vIOBase + r))
#endif /* MUST BYTESWAP */

#define PM3_SAVE_REG(r) l_fb_info->pm3_regs[r >> 3] = PM3_READ_REG(r)
#define PM3_RESTORE_REG(r) PM3_SLOW_WRITE_REG(r, l_fb_info->pm3_regs[r >> 3])

#define PM3_STORE_REG(r,v) l_fb_info->pm3_regs[r >> 3] = v

#define PM3_STORE_DAC_REG(r, v) l_fb_info->dac_regs[r] = v

#define PM3_RESTORE_DAC_REG(r) \
do { \
     DASSERT((l_fb_info->vIOBase != (unsigned char*)(-1)), "l_fb_info->vIOBase mapped in restore dac reg\n"); \
     PM3_SET_INDEX(r); \
     mb(); \
     PM3_WRITE_REG(PM3RD_IndexedData, l_fb_info->dac_regs[r]); \
} while (0)


/* ************************************** */
/* ***** permedia-specific function ***** */
/* ************************************** */

/* Calculating various clock parameter */
static unsigned long pm3fb_CalculateClock(struct pm3fb_info *l_fb_info,
					  unsigned long reqclock,		/* In kHz units */
					  unsigned long refclock,		/* In kHz units */
					  unsigned char *prescale,		/* ClkPreScale */
					  unsigned char *feedback, 		/* ClkFeedBackScale */
					  unsigned char *postscale		/* ClkPostScale */
											)
{
    int			f, pre, post;
    unsigned long	freq;
    long		freqerr = 1000;
    unsigned long  	actualclock = 0;

    DTRACE;

    for (f=1;f<256;f++)
      {
	for (pre=1;pre<256;pre++)
	  {
	    for (post=0;post<5;post++)
	      { 
		freq = ((2* refclock * f) / (pre * (1 << post)));
		if ((reqclock > freq - freqerr)&&(reqclock < freq + freqerr))
		  {
		    freqerr = (reqclock > freq) ? 
		      reqclock - freq : freq - reqclock;
		    *feedback = f;
		    *prescale = pre;
		    *postscale = post;
		    actualclock = freq;
		  }
	      }
	  }
      }
    
    return(actualclock);
}

/*
 * partial products for the supported horizontal resolutions.
 * from pm2fb
 */
#define PACKPP(p0,p1,p2)	(((p2)<<6)|((p1)<<3)|(p0))
static const struct {
	u16 width;
	u16 pp;
} pp_table[] = {
	{ 32,	PACKPP(1, 0, 0) }, { 64,	PACKPP(1, 1, 0) },
	{ 96,	PACKPP(1, 1, 1) }, { 128,	PACKPP(2, 1, 1) },
	{ 160,	PACKPP(2, 2, 1) }, { 192,	PACKPP(2, 2, 2) },
	{ 224,	PACKPP(3, 2, 1) }, { 256,	PACKPP(3, 2, 2) },
	{ 288,	PACKPP(3, 3, 1) }, { 320,	PACKPP(3, 3, 2) },
	{ 384,	PACKPP(3, 3, 3) }, { 416,	PACKPP(4, 3, 1) },
	{ 448,	PACKPP(4, 3, 2) }, { 512,	PACKPP(4, 3, 3) },
	{ 544,	PACKPP(4, 4, 1) }, { 576,	PACKPP(4, 4, 2) },
	{ 640,	PACKPP(4, 4, 3) }, { 768,	PACKPP(4, 4, 4) },
	{ 800,	PACKPP(5, 4, 1) }, { 832,	PACKPP(5, 4, 2) },
	{ 896,	PACKPP(5, 4, 3) }, { 1024,	PACKPP(5, 4, 4) },
	{ 1056,	PACKPP(5, 5, 1) }, { 1088,	PACKPP(5, 5, 2) },
	{ 1152,	PACKPP(5, 5, 3) }, { 1280,	PACKPP(5, 5, 4) },
	{ 1536,	PACKPP(5, 5, 5) }, { 1568,	PACKPP(6, 5, 1) },
	{ 1600,	PACKPP(6, 5, 2) }, { 1664,	PACKPP(6, 5, 3) },
	{ 1792,	PACKPP(6, 5, 4) }, { 2048,	PACKPP(6, 5, 5) },
	{ 0,	0 } };

#define PICOS2KHZ(a) (1000000000UL/(a))
#define KHZ2PICOS(a) (1000000000UL/(a))

static u32 partprod(struct pm3fb_info *l_fb_info, u32 xres)
{
  int i;
  
  DTRACE;
    
  for (i=0; pp_table[i].width && pp_table[i].width!=xres; i++);
  if (!pp_table[i].width)
    DPRINTK(1, "invalid width %u\n", xres);
  return pp_table[i].pp;
}

static int pm3fb_Shiftbpp(struct pm3fb_info *l_fb_info, int v)
{
  DTRACE;

  switch (current_par.depth)
    {
    case 8:
      l_fb_info->BppShift = 4;
      return(v >> 4);
    case 16:
      l_fb_info->BppShift = 3;
      return(v >> 3);
    case 32:
      l_fb_info->BppShift = 2;
      return(v >> 2);
    }
  DPRINTK(1, "Unsupported depth\n");
  return(0);
}

/*
static int Unshiftbpp(struct pm3fb_info *l_fb_info, int v)
{
  DTRACE;

  switch (current_par.depth)
    {
    case 8:
      l_fb_info->BppShift = 4;
      return(v << 4);
    case 16:
      l_fb_info->BppShift = 3;
      return(v << 3);
    case 32:
      l_fb_info->BppShift = 2;
      return(v << 2);
    }
  DPRINTK("Unsupported depth\n");
  return(0);
}
*/

static void pm3_mapIO(struct pm3fb_info *l_fb_info)
{
  DTRACE;

  l_fb_info->vIOBase = ioremap((unsigned long)l_fb_info->pIOBase, 0x20000);
  l_fb_info->v_fb = ioremap((unsigned long)l_fb_info->p_fb, l_fb_info->fb_size);
  DPRINTK(2, "IO mapping : IOBase %lx / %lx, fb %lx / %lx\n",
	  (unsigned long)l_fb_info->pIOBase,
	  (unsigned long)l_fb_info->vIOBase,
	  (unsigned long)l_fb_info->p_fb,
	  (unsigned long)l_fb_info->v_fb);
}

static void pm3_unmapIO(struct pm3fb_info *l_fb_info)
{
  DTRACE;

  iounmap(l_fb_info->vIOBase);
  iounmap(l_fb_info->v_fb);
  l_fb_info->vIOBase = (unsigned char *)-1;
  l_fb_info->v_fb = (unsigned char *)-1;
}

#if defined(PM3FB_MASTER_DEBUG) && (PM3FB_MASTER_DEBUG >= 2)
#define PM3_SHOW_CUR_MODE pm3_show_cur_mode(l_fb_info);
#define PM3_SHOW_STO_MODE pm3_show_sto_mode(l_fb_info);
#define PM3_SHOW_CUR_TIMING pm3_show_cur_timing(l_fb_info);
static void pm3_show_cur_mode(struct pm3fb_info *l_fb_info)
{
  DPRINTK(2, "PM3Aperture0: 0x%08x\n", PM3_READ_REG(PM3Aperture0));
  DPRINTK(2, "PM3Aperture1: 0x%08x\n", PM3_READ_REG(PM3Aperture1));
  DPRINTK(2, "PM3ByAperture1Mode: 0x%08x\n", PM3_READ_REG(PM3ByAperture1Mode));
  DPRINTK(2, "PM3ByAperture2Mode: 0x%08x\n", PM3_READ_REG(PM3ByAperture2Mode));
  DPRINTK(2, "PM3ChipConfig: 0x%08x\n", PM3_READ_REG(PM3ChipConfig));
  DPRINTK(2, "PM3FIFODis: 0x%08x\n", PM3_READ_REG(PM3FIFODis));
  DPRINTK(2, "PM3HTotal: 0x%08x\n", PM3_READ_REG(PM3HTotal));
  DPRINTK(2, "PM3HbEnd: 0x%08x\n", PM3_READ_REG(PM3HbEnd));
  DPRINTK(2, "PM3HgEnd: 0x%08x\n", PM3_READ_REG(PM3HgEnd));
  DPRINTK(2, "PM3HsEnd: 0x%08x\n", PM3_READ_REG(PM3HsEnd));
  DPRINTK(2, "PM3HsStart: 0x%08x\n", PM3_READ_REG(PM3HsStart));
  DPRINTK(2, "PM3MemBypassWriteMask: 0x%08x\n", PM3_READ_REG(PM3MemBypassWriteMask));
  DPRINTK(2, "PM3RD_IndexControl: 0x%08x\n", PM3_READ_REG(PM3RD_IndexControl));
  DPRINTK(2, "PM3ScreenBase: 0x%08x\n", PM3_READ_REG(PM3ScreenBase));
  DPRINTK(2, "PM3ScreenStride: 0x%08x\n", PM3_READ_REG(PM3ScreenStride));
  DPRINTK(2, "PM3VClkCtl: 0x%08x\n", PM3_READ_REG(PM3VClkCtl));
  DPRINTK(2, "PM3VTotal: 0x%08x\n", PM3_READ_REG(PM3VTotal));
  DPRINTK(2, "PM3VbEnd: 0x%08x\n", PM3_READ_REG(PM3VbEnd));
  DPRINTK(2, "PM3VideoControl: 0x%08x\n", PM3_READ_REG(PM3VideoControl));
  DPRINTK(2, "PM3VsEnd: 0x%08x\n", PM3_READ_REG(PM3VsEnd));
  DPRINTK(2, "PM3VsStart: 0x%08x\n", PM3_READ_REG(PM3VsStart));
}

static void pm3_show_sto_mode(struct pm3fb_info *l_fb_info)
{
  DPRINTK(2, "PM3Aperture0: 0x%08lx\n", l_fb_info->pm3_regs[PM3Aperture0 >> 3]);
  DPRINTK(2, "PM3Aperture1: 0x%08lx\n", l_fb_info->pm3_regs[PM3Aperture1 >> 3]);
  DPRINTK(2, "PM3ByAperture1Mode: 0x%08lx\n", l_fb_info->pm3_regs[PM3ByAperture1Mode >> 3]);
  DPRINTK(2, "PM3ByAperture2Mode: 0x%08lx\n", l_fb_info->pm3_regs[PM3ByAperture2Mode >> 3]);
  DPRINTK(2, "PM3ChipConfig: 0x%08lx\n", l_fb_info->pm3_regs[PM3ChipConfig >> 3]);
  DPRINTK(2, "PM3FIFODis: 0x%08lx\n", l_fb_info->pm3_regs[PM3FIFODis >> 3]);
  DPRINTK(2, "PM3HTotal: 0x%08lx\n", l_fb_info->pm3_regs[PM3HTotal >> 3]);
  DPRINTK(2, "PM3HbEnd: 0x%08lx\n", l_fb_info->pm3_regs[PM3HbEnd >> 3]);
  DPRINTK(2, "PM3HgEnd: 0x%08lx\n", l_fb_info->pm3_regs[PM3HgEnd >> 3]);
  DPRINTK(2, "PM3HsEnd: 0x%08lx\n", l_fb_info->pm3_regs[PM3HsEnd >> 3]);
  DPRINTK(2, "PM3HsStart: 0x%08lx\n", l_fb_info->pm3_regs[PM3HsStart >> 3]);
  DPRINTK(2, "PM3MemBypassWriteMask: 0x%08lx\n", l_fb_info->pm3_regs[PM3MemBypassWriteMask >> 3]);
  DPRINTK(2, "PM3RD_IndexControl: 0x%08lx\n", l_fb_info->pm3_regs[PM3RD_IndexControl >> 3]);
  DPRINTK(2, "PM3ScreenBase: 0x%08lx\n", l_fb_info->pm3_regs[PM3ScreenBase >> 3]);
  DPRINTK(2, "PM3ScreenStride: 0x%08lx\n", l_fb_info->pm3_regs[PM3ScreenStride >> 3]);
  DPRINTK(2, "PM3VClkCtl: 0x%08lx\n", l_fb_info->pm3_regs[PM3VClkCtl >> 3]);
  DPRINTK(2, "PM3VTotal: 0x%08lx\n", l_fb_info->pm3_regs[PM3VTotal >> 3]);
  DPRINTK(2, "PM3VbEnd: 0x%08lx\n", l_fb_info->pm3_regs[PM3VbEnd >> 3]);
  DPRINTK(2, "PM3VideoControl: 0x%08lx\n", l_fb_info->pm3_regs[PM3VideoControl >> 3]);
  DPRINTK(2, "PM3VsEnd: 0x%08lx\n", l_fb_info->pm3_regs[PM3VsEnd >> 3]);
  DPRINTK(2, "PM3VsStart: 0x%08lx\n", l_fb_info->pm3_regs[PM3VsStart >> 3]);
}

static void  pm3_show_cur_timing(struct pm3fb_info *l_fb_info)
{
  DPRINTK(2, "PM3LocalMemCaps: 0x%08x\n", PM3_READ_REG(PM3LocalMemCaps));
  DPRINTK(2, "PM3LocalMemTimings: 0x%08x\n", PM3_READ_REG(PM3LocalMemTimings));
  DPRINTK(2, "PM3LocalMemControl: 0x%08x\n", PM3_READ_REG(PM3LocalMemControl));
  DPRINTK(2, "PM3LocalMemRefresh: 0x%08x\n", PM3_READ_REG(PM3LocalMemRefresh));
  DPRINTK(2, "PM3LocalMemPowerDown: 0x%08x\n", PM3_READ_REG(PM3LocalMemPowerDown));
}
#else
#define PM3_SHOW_CUR_MODE /* pm3_show_mode(); */
#define PM3_SHOW_STO_MODE /* pm3_show_sto_mode(); */
#define PM3_SHOW_CUR_TIMING /* pm3_show_cur_timing(); */
#endif

/* Store the current mode (aka current_par) in the local shadow register */
static void pm3_store_mode(struct pm3fb_info *l_fb_info)
{
  DTRACE;

  PM3_STORE_REG(PM3MemBypassWriteMask,  	0xffffffff);
  PM3_STORE_REG(PM3Aperture0,		 	0x00000000);
  PM3_STORE_REG(PM3Aperture1,		 	0x00000000);
  PM3_STORE_REG(PM3FIFODis,			0x00000007);

  PM3_STORE_REG(PM3HTotal, pm3fb_Shiftbpp(l_fb_info, current_par.htotal - 1));
  PM3_STORE_REG(PM3HsEnd, pm3fb_Shiftbpp(l_fb_info, current_par.hsend));
  PM3_STORE_REG(PM3HsStart, pm3fb_Shiftbpp(l_fb_info, current_par.hsstart));
  PM3_STORE_REG(PM3HbEnd, pm3fb_Shiftbpp(l_fb_info, current_par.hbend));
  PM3_STORE_REG(PM3HgEnd, pm3fb_Shiftbpp(l_fb_info, current_par.hbend));
  PM3_STORE_REG(PM3ScreenStride, pm3fb_Shiftbpp(l_fb_info, current_par.stride));
  PM3_STORE_REG(PM3VTotal, current_par.vtotal - 1);
  PM3_STORE_REG(PM3VsEnd, current_par.vsend - 1);
  PM3_STORE_REG(PM3VsStart, current_par.vsstart - 1);
  PM3_STORE_REG(PM3VbEnd, current_par.vbend);

  switch (current_par.depth)
    {
    case 8:
      PM3_STORE_REG(PM3ByAperture1Mode, PM3ByApertureMode_PIXELSIZE_8BIT);
      PM3_STORE_REG(PM3ByAperture2Mode, PM3ByApertureMode_PIXELSIZE_8BIT);
	  /*
      PM3_STORE_REG(PM3VideoControl, PM3VideoControl_ENABLE |
		    PM3VideoControl_HSYNC_ACTIVE_HIGH |
		    PM3VideoControl_VSYNC_ACTIVE_HIGH |
		    PM3VideoControl_PIXELSIZE_8BIT);
	  */
	  
      break;
    case 16:
#ifdef __BIG_ENDIAN
      PM3_STORE_REG(PM3ByAperture1Mode, PM3ByApertureMode_PIXELSIZE_16BIT);
      PM3_STORE_REG(PM3ByAperture2Mode, PM3ByApertureMode_PIXELSIZE_16BIT);
#else
      PM3_STORE_REG(PM3ByAperture1Mode, PM3ByApertureMode_PIXELSIZE_16BIT |
					PM3ByApertureMode_BYTESWAP_BADC);
      PM3_STORE_REG(PM3ByAperture2Mode, PM3ByApertureMode_PIXELSIZE_16BIT |
					PM3ByApertureMode_BYTESWAP_BADC);
#endif
	  /*
	  PM3_STORE_REG(PM3VideoControl, PM3VideoControl_ENABLE |
					PM3VideoControl_HSYNC_ACTIVE_HIGH |
					PM3VideoControl_VSYNC_ACTIVE_HIGH |
					PM3VideoControl_PIXELSIZE_16BIT);
	  */
      break;
    case 32:
#ifdef __BIG_ENDIAN
      PM3_STORE_REG(PM3ByAperture1Mode, PM3ByApertureMode_PIXELSIZE_32BIT);
      PM3_STORE_REG(PM3ByAperture2Mode, PM3ByApertureMode_PIXELSIZE_32BIT);
#else
      PM3_STORE_REG(PM3ByAperture1Mode, PM3ByApertureMode_PIXELSIZE_32BIT |
		    PM3ByApertureMode_BYTESWAP_DCBA);
      PM3_STORE_REG(PM3ByAperture2Mode, PM3ByApertureMode_PIXELSIZE_32BIT |
		    PM3ByApertureMode_BYTESWAP_DCBA);
#endif
	  /*
      PM3_STORE_REG(PM3VideoControl, PM3VideoControl_ENABLE |
		    PM3VideoControl_HSYNC_ACTIVE_HIGH |
		    PM3VideoControl_VSYNC_ACTIVE_HIGH |
		    PM3VideoControl_PIXELSIZE_32BIT);
	  */
      break;
    }

  PM3_STORE_REG(PM3VideoControl, current_par.video);
  PM3_STORE_REG(PM3VClkCtl, (PM3_READ_REG(PM3VClkCtl) & 0xFFFFFFFC));
  PM3_STORE_REG(PM3ScreenBase, 0x00000000);
  PM3_STORE_REG(PM3ChipConfig, (PM3_READ_REG(PM3ChipConfig) & 0xFFFFFFFD));
  
  {
    unsigned char m;		/* ClkPreScale */
    unsigned char n; 		/* ClkFeedBackScale */
    unsigned char p;		/* ClkPostScale */
    unsigned long usedclock = pm3fb_CalculateClock(l_fb_info, current_par.pixclock, 14318, &m, &n, &p); /* usedclock unused */

    PM3_STORE_DAC_REG(PM3RD_DClk0PreScale, m);
    PM3_STORE_DAC_REG(PM3RD_DClk0FeedbackScale, n);
    PM3_STORE_DAC_REG(PM3RD_DClk0PostScale, p);
  }
  /*
  PM3_STORE_DAC_REG(PM3RD_IndexControl, 0x00);
  */
  /*
  PM3_STORE_REG(PM3RD_IndexControl, 0x00);
  */
  {
    char tempsync = 0x00;
    
    if ((current_par.video & PM3VideoControl_HSYNC_MASK) == PM3VideoControl_HSYNC_ACTIVE_HIGH)
      tempsync |= PM3RD_SyncControl_HSYNC_ACTIVE_HIGH;
    if ((current_par.video & PM3VideoControl_VSYNC_MASK) == PM3VideoControl_VSYNC_ACTIVE_HIGH)
      tempsync |= PM3RD_SyncControl_VSYNC_ACTIVE_HIGH;
    
    PM3_STORE_DAC_REG(PM3RD_SyncControl, tempsync);
	DPRINTK(1, "PM3RD_SyncControl: %d\n", tempsync);
  }
  PM3_STORE_DAC_REG(PM3RD_DACControl, 0x00);
  
  switch (current_par.depth)
    {
    case 8:
      PM3_STORE_DAC_REG(PM3RD_PixelSize, PM3RD_PixelSize_8_BIT_PIXELS);
      PM3_STORE_DAC_REG(PM3RD_ColorFormat, 0x2E);
      break;
    case 16:
      PM3_STORE_DAC_REG(PM3RD_PixelSize, PM3RD_PixelSize_16_BIT_PIXELS);
      PM3_STORE_DAC_REG(PM3RD_ColorFormat, 0x70);
      break;
    case 32:
      PM3_STORE_DAC_REG(PM3RD_PixelSize, PM3RD_PixelSize_32_BIT_PIXELS);
      PM3_STORE_DAC_REG(PM3RD_ColorFormat, 0x20);
      break;
    }

  PM3_STORE_DAC_REG(PM3RD_MiscControl, PM3RD_MiscControl_HIGHCOLOR_RES_ENABLE);

  PM3_SHOW_STO_MODE  
}

/* put the shoadow register in the real hardware */
static void pm3_restore_mode(struct pm3fb_info *l_fb_info)
{
  DTRACE;

  DASSERT((l_fb_info->vIOBase != (unsigned char*)-1), "IOBase mapped\n");

  PM3_RESTORE_REG(PM3MemBypassWriteMask);
  PM3_RESTORE_REG(PM3Aperture0);
  PM3_RESTORE_REG(PM3Aperture1);
  PM3_RESTORE_REG(PM3FIFODis);

  PM3_RESTORE_REG(PM3VideoControl);
  PM3_RESTORE_REG(PM3ByAperture1Mode);
  PM3_RESTORE_REG(PM3ByAperture2Mode);

  PM3_RESTORE_REG(PM3HbEnd);
  PM3_RESTORE_REG(PM3HgEnd);
  PM3_RESTORE_REG(PM3ScreenStride);
  PM3_RESTORE_REG(PM3HTotal);
  PM3_RESTORE_REG(PM3HsStart);
  PM3_RESTORE_REG(PM3HsEnd);
  PM3_RESTORE_REG(PM3VTotal);
  PM3_RESTORE_REG(PM3VbEnd);
  PM3_RESTORE_REG(PM3VsStart);
  PM3_RESTORE_REG(PM3VsEnd);
    
  PM3_RESTORE_REG(PM3ScreenBase);
  PM3_RESTORE_REG(PM3VClkCtl);
  PM3_RESTORE_REG(PM3ChipConfig);
 
  PM3_RESTORE_DAC_REG(PM3RD_DClk0FeedbackScale);
  PM3_RESTORE_DAC_REG(PM3RD_DClk0PostScale);
  PM3_RESTORE_DAC_REG(PM3RD_DClk0PreScale);
  
  /*
  PM3_RESTORE_DAC_REG(PM3RD_IndexControl);
  */
  /*
  PM3_RESTORE_REG(PM3RD_IndexControl);
  */
  PM3_RESTORE_DAC_REG(PM3RD_SyncControl);
  PM3_RESTORE_DAC_REG(PM3RD_DACControl);
  PM3_RESTORE_DAC_REG(PM3RD_PixelSize);
  PM3_RESTORE_DAC_REG(PM3RD_ColorFormat);
  PM3_RESTORE_DAC_REG(PM3RD_MiscControl);

  PM3_SHOW_CUR_MODE
}


static unsigned long pm3_size_memory(struct pm3fb_info *l_fb_info)
{
  unsigned long memsize, tempBypass, i, temp1, temp2;
  
  DTRACE;

#if 0
  memsize =0x1000000; /* force 16 MB */
#else
  l_fb_info->fb_size = 64*1024*1024; /* pm3 aperture alwayx 64 MB */
  pm3_mapIO(l_fb_info); /* temporary map IO */
  
  DASSERT((l_fb_info->vIOBase != NULL), "IO successfully mapped before mem detect\n");
  DASSERT((l_fb_info->v_fb != NULL), "FBsuccessfully mapped before mem detect\n");

  tempBypass = PM3_READ_REG(PM3MemBypassWriteMask);

  DPRINTK(2, "PM3MemBypassWriteMask was: 0x%08lx", tempBypass);

  PM3_SLOW_WRITE_REG(PM3MemBypassWriteMask, 0xFFFFFFFF);

  /* pm3 split up memory, replicates, and do a lot pof nasty stuff IMHO ;-) */

  for(i = 0; i < 32; i++)
	{
#ifdef MUST_BYTESWAP
	  writel(__swab32(i * 0x00345678), (l_fb_info->v_fb + (i * 1048576)));
#else
	  writel(i * 0x00345678, (l_fb_info->v_fb + (i * 1048576)));
#endif
	  mb();
#ifdef MUST_BYTESWAP
	  temp1 = __swab32(readl((l_fb_info->v_fb + (i * 1048576))));
#else
	  temp1 = readl((l_fb_info->v_fb + (i * 1048576)));
#endif
	  /* Let's check for wrapover, write will fail at 16MB boundary */
	  if (temp1 == (i * 0x00345678)) 
		memsize = i;
	  else 
		break;
	}

  DPRINTK(2, "First detect pass already got %ld MB\n", memsize);

  /* ok we've done the first 32 mb... */
  
  if (memsize == i)
	{
	  for(i=0;i<32;i++)
		{
		  /* Clear first 32MB ; 0 is 0, no need to byteswap */
		  writel(0x0000000, (l_fb_info->v_fb + (i * 1048576)));
		  mb();
		}

	  for(i=32;i<64;i++)
		{
#ifdef MUST_BYTESWAP
		  writel(__swab32(i * 0x00345678), (l_fb_info->v_fb + (i * 1048576)));
#else
		  writel(i * 0x00345678, (l_fb_info->v_fb + (i * 1048576)));
#endif
		  mb();
#ifdef MUST_BYTESWAP
		  temp1 = __swab32(readl((l_fb_info->v_fb + (i * 1048576))));
		  temp2 = __swab32(readl((l_fb_info->v_fb + ((i - 32) * 1048576))));
#else
		  temp1 = readl((l_fb_info->v_fb + (i * 1048576)));
		  temp2 = readl((l_fb_info->v_fb + ((i - 32) * 1048576)));
#endif
		  if ( (temp1 == (i*0x00345678)) && (temp2 == 0) ) /* different value, different RAM... */
			memsize = i;
		  else 
			break;
		}
	}

  DPRINTK(2, "Second detect pass got %ld MB\n", memsize);

  PM3_SLOW_WRITE_REG(PM3MemBypassWriteMask, tempBypass);

  pm3_unmapIO(l_fb_info);
  memsize = 1048576 * (memsize + 1);
  
  DPRINTK(2, "Returning 0x%08lx bytes\n", memsize);

  l_fb_info->fb_size = memsize;
#endif
  return memsize;
}

/* ********************************************** */
/* ***** framebuffer API standard functions ***** */
/* ********************************************** */

static int pm3fb_open(struct fb_info* info, int user)
{
  struct pm3fb_info *l_fb_info = (struct pm3fb_info*) info;

  DTRACE;

  DPRINTK(3, "user is %d\n", user);

  MOD_INC_USE_COUNT;
 
  return(0);
}

static int pm3fb_release(struct fb_info* info, int user)
{
  struct pm3fb_info *l_fb_info = (struct pm3fb_info*) info;

  DTRACE;

  MOD_DEC_USE_COUNT;

  return(0);
}

static int pm3fb_encode_fix(struct fb_fix_screeninfo *fix,
			    const void *par,
			    struct fb_info_gen *info)
{
  struct pm3fb_info* l_fb_info=(struct pm3fb_info* )info;
  struct pm3fb_par* p=(struct pm3fb_par* )par;

  DTRACE;
  
  strcpy(fix->id, permedia3_name);
  fix->smem_start = l_fb_info->p_fb;
  fix->smem_len = l_fb_info->fb_size;
  fix->mmio_start = l_fb_info->pIOBase;
  fix->mmio_len = PM3_REGS_SIZE;
  fix->accel = 0 /* FB_ACCEL_3DLABS_PERMEDIA3 */;
  fix->type = FB_TYPE_PACKED_PIXELS;
  fix->visual= (p->depth == 8) ? FB_VISUAL_PSEUDOCOLOR:FB_VISUAL_TRUECOLOR;
  if (current_par_valid)
	fix->line_length = current_par.width*(current_par.depth/8);
  else
	fix->line_length = 0;
  fix->xpanstep = 64/p->depth;
  fix->ypanstep = 1;
  fix->ywrapstep = 0;
  return 0;
}



static int pm3fb_decode_var(const struct fb_var_screeninfo* var,
			    void *par,
			    struct fb_info_gen* info)
{
  struct pm3fb_info *l_fb_info = (struct pm3fb_info*)info;
  struct pm3fb_par *p = (struct pm3fb_par*)par;
  u32 xres;
  
  DTRACE;

  DASSERT((var != NULL), "fb_var_screeninfo* not NULL");
  DASSERT((p != NULL), "pm3fb_par* not NULL");
  DASSERT((l_fb_info != NULL), "pm3fb_info* not NULL");

  memset(p, 0, sizeof(struct pm3fb_par));
  p->width = (var->xres_virtual + 7) & ~7;
  p->height = var->yres_virtual;
  p->depth = (var->bits_per_pixel + 7) & ~7;
  p->depth = (p->depth > 32) ? 32 : p->depth;

  xres = (var->xres + 31) & ~31;
  if (p->width<xres+var->xoffset)
    p->width=xres+var->xoffset;
  if (p->height<var->yres+var->yoffset)
    p->height=var->yres+var->yoffset;

  DPRINTK(3, "First var filling done\n");
  
  
  if (!partprod(l_fb_info, xres)) {
    DPRINTK(1, "width not supported: %u\n", xres);
    return -EINVAL;
  }
  
  if (p->width>2047) {
    DPRINTK(1, "virtual width not supported: %u\n", p->width);
    return -EINVAL;
  }
  if (var->yres<200) {
    DPRINTK(1, "height not supported: %u\n",
	    (u32 )var->yres);
    return -EINVAL;
  }
  if (p->height<200 || p->height>2047) {
    DPRINTK(1, "virtual height not supported: %u\n", p->height);
    return -EINVAL;
  }
  if (p->depth>32) {
    DPRINTK(1, "depth not supported: %u\n", p->depth);
    return -EINVAL;
  }
  if (p->width*p->height*p->depth/8>l_fb_info->fb_size) {
    DPRINTK(1, "no memory for screen (%ux%ux%u)\n",
	    p->width, p->height, p->depth);
    return -EINVAL;
  }
  p->pixclock=PICOS2KHZ(var->pixclock);
  if (p->pixclock>PM3_MAX_PIXCLOCK) {
    DPRINTK(1, "pixclock too high (%uKHz)\n", p->pixclock);
    return -EINVAL;
  }

  DPRINTK(3, "Clock done\n");

  p->hsstart = var->right_margin;
  p->hsend = var->right_margin + var->hsync_len;
  p->hbend = var->right_margin + var->hsync_len + var->left_margin;
  p->htotal = xres + p->hbend;

  p->vsstart = var->lower_margin;
  p->vsend = var->lower_margin + var->vsync_len;
  p->vbend = var->lower_margin + var->vsync_len + var->upper_margin;
  p->vtotal = var->yres + p->vbend;

  p->stride = p->width;

  p->base = (var->yoffset * xres) + var->xoffset;
  
  p->video = 0;
  
  if (var->sync & FB_SYNC_HOR_HIGH_ACT)
    p->video |= PM3VideoControl_HSYNC_ACTIVE_HIGH;
  else
    p->video |= PM3VideoControl_HSYNC_ACTIVE_LOW;
  
  if (var->sync & FB_SYNC_VERT_HIGH_ACT)
    p->video |= PM3VideoControl_VSYNC_ACTIVE_HIGH;
  else
    p->video |= PM3VideoControl_VSYNC_ACTIVE_LOW;
  
  if ((var->vmode & FB_VMODE_MASK) == FB_VMODE_INTERLACED)
    {
      DPRINTK(1, "interlaced not supported\n\n");
      return -EINVAL;
    }
  
  if ((var->vmode & FB_VMODE_MASK) == FB_VMODE_DOUBLE)
    p->video |= PM3VideoControl_LINE_DOUBLE_ON;
  else
    p->video |= PM3VideoControl_LINE_DOUBLE_OFF;
  
  if (var->activate == FB_ACTIVATE_NOW)
    p->video |= PM3VideoControl_ENABLE;
  else
    {
      p->video |= PM3VideoControl_DISABLE;
      DPRINTK(2, "PM3Video disabled\n");
    }

  switch (p->depth)
    {
    case 8:
      p->video |= PM3VideoControl_PIXELSIZE_8BIT;
      break;
    case 16:
      p->video |= PM3VideoControl_PIXELSIZE_16BIT;
      break;
    case 32:
      p->video |= PM3VideoControl_PIXELSIZE_32BIT;
      break;
    default:
      DPRINTK(1, "Unsupported depth\n");
      break;
    }

  DPRINTK(3, "All done\n");
  
  return 0;
}

static int pm3fb_encode_var(struct fb_var_screeninfo *var,
			    const void *par,
			    struct fb_info_gen *info)
{
  struct pm3fb_par* p= (struct pm3fb_par*) par;
  struct pm3fb_info *l_fb_info = (struct pm3fb_info*) info;

  u32 base;
  
  DTRACE; 

  DASSERT((var != NULL), "fb_var_screeninfo* not NULL");
  DASSERT((p != NULL), "pm3fb_par* not NULL");
  DASSERT((info != NULL), "fb_info_gen* not NULL");
 
  memset(var, 0, sizeof(struct fb_var_screeninfo));
  /*
    if (p->flags & PM2FF_ACCEL)
    var->accel_flags |= FB_ACCELF_TEXT;
  */
  var->xres_virtual = p->width;
  var->yres_virtual = p->height;
  var->xres = p->htotal - p->hbend;
  var->yres = p->vtotal - p->vbend;

  DPRINTK(2, "xres = %d, yres : %d\n", var->xres, var->yres);

  var->right_margin = p->hsstart;
  var->hsync_len = p->hsend - p->hsstart;
  var->left_margin = p->hbend - p->hsend;
  var->lower_margin = p->vsstart;
  var->vsync_len = p->vsend - p->vsstart;
  var->upper_margin = p->vbend - p->vsend;
  var->bits_per_pixel = p->depth;
  
  switch (p->depth) {
  case 8:
    var->red.length = var->green.length = var->blue.length = 8;
    /*
    var->xres = var->xres<<4;
    var->right_margin = var->right_margin<<4;
    var->hsync_len = var->hsync_len<<4;
    var->left_margin = var->left_margin<<4;
    */
    break;
  case 16:
    var->red.offset = 11;
    var->red.length = 5;
    var->green.offset = 5;
    var->green.length = 6;
    var->blue.length = 5;
    /*
    var->xres = var->xres<<3;
    var->right_margin = var->right_margin<<3;
    var->hsync_len = var->hsync_len<<3;
    var->left_margin = var->left_margin<<3;
    */
    break;
  case 32:
    var->transp.offset = 24;
    var->red.offset = 16;
    var->green.offset = 8;
    /*
    var->xres = var->xres<<2;
    var->right_margin = var->right_margin<<2;
    var->hsync_len = var->hsync_len<<2;
    var->left_margin = var->left_margin<<2;
    */
    var->red.length = var->green.length = var->blue.length = var->transp.length = 8;
    break;
  default:
    DPRINTK(1, "Unsupported depth\n");
    break;
  }

  base = p->base;

  var->xoffset = base % var->xres;
  var->yoffset = base / var->xres;

  var->height = var->width = -1;

  var->pixclock = KHZ2PICOS(p->pixclock);

  DPRINTK(3, "Clock done\n");

  if ((p->video & PM3VideoControl_HSYNC_MASK) == PM3VideoControl_HSYNC_ACTIVE_HIGH)
    var->sync |= FB_SYNC_HOR_HIGH_ACT;
  if ((p->video & PM3VideoControl_VSYNC_MASK) == PM3VideoControl_VSYNC_ACTIVE_HIGH)
    var->sync |= FB_SYNC_VERT_HIGH_ACT;
  if (p->video & PM3VideoControl_LINE_DOUBLE_ON)
    var->vmode = FB_VMODE_DOUBLE;

  DPRINTK(3, "all done\n");

  return 0;
}

static void pm3fb_get_par(void *par, struct fb_info_gen *info)
{
  struct pm3fb_info *l_fb_info = (struct pm3fb_info*) info;

  DTRACE;

  if (!current_par_valid) {
    memcpy(&current_par, &(mode_base[current_mode].user_mode), sizeof(current_par));
    pm3_store_mode(l_fb_info);
    current_par_valid = 1;
  }
  *((struct pm3fb_par* )par) = current_par;
}

static void pm3fb_set_par(const void *par, struct fb_info_gen *info)
{
  struct pm3fb_info *l_fb_info = (struct pm3fb_info*) info;

  DTRACE;
  
  current_par  =  *((struct pm3fb_par*)par);
  current_par_valid = 1;
  pm3_store_mode(l_fb_info);
  pm3_restore_mode(l_fb_info);
}

static void pm3fb_set_color(struct pm3fb_info* l_fb_info, unsigned char regno,
			unsigned char r, unsigned char g, unsigned char b)
{
  DTRACE;

  DPRINTK(3, "Writing color #%d, rgb %3d:%3d:%3d\n", (int)regno, (int)r, (int)g, (int)b);

  PM3_SLOW_WRITE_REG(PM3RD_PaletteWriteAddress, regno);
  PM3_SLOW_WRITE_REG(PM3RD_PaletteData, r);
  PM3_SLOW_WRITE_REG(PM3RD_PaletteData, g);
  PM3_SLOW_WRITE_REG(PM3RD_PaletteData, b);
}

static int pm3fb_getcolreg(unsigned regno, unsigned *red, unsigned *green,
			   unsigned *blue, unsigned *transp,
			   struct fb_info *info)
{
  struct pm3fb_info* l_fb_info = (struct pm3fb_info* )info;
  
  DTRACE;

  if (regno<256) {
    *red = l_fb_info->palette[regno].red<<8|l_fb_info->palette[regno].red;
    *green = l_fb_info->palette[regno].green<<8|l_fb_info->palette[regno].green;
    *blue = l_fb_info->palette[regno].blue<<8|l_fb_info->palette[regno].blue;
    *transp = l_fb_info->palette[regno].transp<<8|l_fb_info->palette[regno].transp;
  }
  return regno>255;
}

static int pm3fb_setcolreg(unsigned regno, unsigned red, unsigned green,
			   unsigned blue, unsigned transp,
			   struct fb_info *info)
{
  struct pm3fb_info* l_fb_info = (struct pm3fb_info* )info;
  
  DTRACE;
  
  if (regno<16) {
    switch (current_par.depth) {
#ifdef FBCON_HAS_CFB8
    case 8:
      break;
#endif
#ifdef FBCON_HAS_CFB16
    case 16:
      l_fb_info->cmap.cmap16[regno]=
	((u32 )red & 0xf800) |
	(((u32 )green & 0xfc00)>>5) |
	(((u32 )blue & 0xf800)>>11);
      break;
#endif
#ifdef FBCON_HAS_CFB32
    case 32:
      l_fb_info->cmap.cmap32[regno]=
	(((u32 )transp & 0xff00) << 16) |
	(((u32 )red & 0xff00) << 8) |
	(((u32 )green & 0xff00)) |
	(((u32 )blue & 0xff00) >> 8);
      break;
#endif
    default:
      DPRINTK(1, "bad depth %u\n", current_par.depth);
      break;
    }
  }
  if (regno<256) {
    l_fb_info->palette[regno].red=red >> 8;
    l_fb_info->palette[regno].green=green >> 8;
    l_fb_info->palette[regno].blue=blue >> 8;
    l_fb_info->palette[regno].transp=transp >> 8;
    if (current_par.depth==8)
      pm3fb_set_color(l_fb_info, regno, red>>8, green>>8, blue>>8);
  }
  return regno>255;
}

static int pm3fb_blank(int blank_mode, struct fb_info_gen *info)
{
  struct pm3fb_info *l_fb_info = (struct pm3fb_info*) info;

  DTRACE;

  /* TO BE DONE */
#if 0
  struct pm2fb_info* i=(struct pm2fb_info* )info;
  u32 video;
  
  if (!current_par_valid)
    return 1;
  video=current_par.video;
  if (blank_mode>0) {
    switch (blank_mode-1) {
    case VESA_NO_BLANKING: /* FIXME */
      video=video&~(PM2F_VIDEO_ENABLE);
      break;
    case VESA_HSYNC_SUSPEND:
      video=video&~(PM2F_HSYNC_MASK|
		    PM2F_BLANK_LOW);
      break;
    case VESA_VSYNC_SUSPEND:
      video=video&~(PM2F_VSYNC_MASK|
		    PM2F_BLANK_LOW);
      break;
    case VESA_POWERDOWN:
      video=video&~(PM2F_VSYNC_MASK|
		    PM2F_HSYNC_MASK|
		    PM2F_BLANK_LOW);
      break;
    }
  }
  WAIT_FIFO(i, 1);
  pm2_WR(i, PM2R_VIDEO_CONTROL, video);
  return 0;
#endif
  return(1);
}

static void pm3fb_set_disp(const void *par, struct display *disp,
			   struct fb_info_gen *info)
{
  struct pm3fb_info *l_fb_info = (struct pm3fb_info *)info;
  struct pm3fb_par *p = (struct pm3fb_par *)par;
  u32 flags;

  DTRACE;
  
  save_flags(flags); cli();
  disp->screen_base=l_fb_info->v_fb;
  switch (p->depth)
    {
#ifdef FBCON_HAS_CFB8
    case 8:
      disp->dispsw=&fbcon_cfb8;
      break;
#endif
#ifdef FBCON_HAS_CFB16
    case 16:
      disp->dispsw=&fbcon_cfb16;
      disp->dispsw_data=l_fb_info->cmap.cmap16;
      break;
#endif
#ifdef FBCON_HAS_CFB32
    case 32:
      disp->dispsw=&fbcon_cfb32;
      disp->dispsw_data=l_fb_info->cmap.cmap32;
      break;
#endif
    default:
      disp->dispsw=&fbcon_dummy;
      DPRINTK(1, "Invalid depth, using fbcon_dummy\n");
      break;
    }
  restore_flags(flags);
}

/* */
static void pm3fb_detect(void /* struct pm3fb_info *l_fb_info */)
{
  struct pci_dev *dev = NULL;
  struct pm3fb_info *l_fb_info = &(fb_info[detected_board]);
  
  /* should be reworked (multiple boards / chips-perp-board...) */
  
  DTRACE;
  
  dev = pci_find_device(PCI_VENDOR_ID_3DLABS, PCI_DEVICE_ID_3DLABS_PERMEDIA3, dev);
  if (!dev)
    return;
  l_fb_info->dev = dev;
#ifdef CONFIG_FB_OF
  l_fb_info->dn = NULL;
#endif


  DPRINTK(1, "found @%lx Vendor %lx Device %lx ; base @ : %lx - %lx - %lx - %lx - %lx - %lx, rom @ %lx, irq %ld\n",
	  (unsigned long)dev,
	  (unsigned long)dev->vendor,
	  (unsigned long)dev->device,
	  (unsigned long)dev->base_address[0],
	  (unsigned long)dev->base_address[1],
	  (unsigned long)dev->base_address[2],
	  (unsigned long)dev->base_address[3],
	  (unsigned long)dev->base_address[4],
	  (unsigned long)dev->base_address[5],
	  (unsigned long)dev->rom_address,
	  (unsigned long)dev->irq);

  l_fb_info->pIOBase = (unsigned char*)(l_fb_info->dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK);
#ifdef __BIG_ENDIAN
  l_fb_info->pIOBase += PM3_REGS_SIZE;
#endif
  l_fb_info->vIOBase = (unsigned char*)-1;
  l_fb_info->p_fb = (unsigned char*)(l_fb_info->dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK);
  l_fb_info->v_fb = (unsigned char*)-1;
  l_fb_info->fb_size = pm3_size_memory(l_fb_info);

  pm3_mapIO(l_fb_info);

  PM3_SHOW_CUR_MODE

  PM3_SHOW_CUR_TIMING
	
  /*
  pm3fb_get_par(&par, &l_fb_info->gen.info);
  pm3fb_encode_var(&default_var, &par, &l_fb_info->gen.info);
  */
}

/* */
struct fbgen_hwswitch pm3fb_switch = {
    pm3fb_detect, pm3fb_encode_fix, pm3fb_decode_var, pm3fb_encode_var,
    pm3fb_get_par, pm3fb_set_par, pm3fb_getcolreg, pm3fb_setcolreg,
    fbgen_pan_display, pm3fb_blank, pm3fb_set_disp
};

/* */
static struct fb_ops pm3fb_ops = {
    pm3fb_open, pm3fb_release, fbgen_get_fix, fbgen_get_var, fbgen_set_var,
    fbgen_get_cmap, fbgen_set_cmap, fbgen_pan_display, fbgen_ioctl, NULL, NULL
};

/* ****************************************** */
/* ***** standard FB API init functions ***** */
/* ****************************************** */

__initfunc(void pm3fb_setup(char *options))
{
  DTRACE;
  
  return;
}

__initfunc(void pm3fb_init(void))
{
  struct pm3fb_info *l_fb_info = &(fb_info[detected_board]);

  DTRACE;

  memset(l_fb_info, 0, sizeof(struct pm3fb_info));
  
  l_fb_info->gen.fbhw = &pm3fb_switch;
  l_fb_info->gen.fbhw->detect();

  if (!l_fb_info->dev)
	{
	  DPRINTK(1, "No Permedia3 board detected\n");
	  return;
	}

  strcpy(l_fb_info->gen.info.modename, permedia3_name);
  disp.scrollmode=SCROLL_YNOMOVE;
  l_fb_info->gen.parsize=sizeof(struct pm3fb_par);
  l_fb_info->gen.info.changevar = NULL;
  l_fb_info->gen.info.node = -1;
  l_fb_info->gen.info.fbops = &pm3fb_ops;
  l_fb_info->gen.info.disp = &disp;
  l_fb_info->gen.info.switch_con = &fbgen_switch;
  l_fb_info->gen.info.updatevar = &fbgen_update_var; /* */
  l_fb_info->gen.info.blank = &fbgen_blank; /* */
  l_fb_info->gen.info.flags = FBINFO_FLAG_DEFAULT;

  DPRINTK(3, "Var filling done\n");
  
  (void)fbgen_get_var(&disp.var, -1, &l_fb_info->gen.info);
  
  DPRINTK(3, "First fbgen done\n");
  
  (void)fbgen_do_set_var(&disp.var, 1, &l_fb_info->gen);
  
  DPRINTK(3, "Second fbgen done\n");

  fbgen_set_disp(-1, &l_fb_info->gen);

  DPRINTK(3, "Third fbgen done\n");

  fbgen_install_cmap(0, &l_fb_info->gen);

  DPRINTK(3, "Fourth fbgen done\n");

  if (register_framebuffer(&l_fb_info->gen.info) < 0)
	{
	  DPRINTK(1, "Couldn't register framebuffer\n");
	  return;
	}

  DPRINTK(3, "Register fb done\n");
  
  pm3_restore_mode(l_fb_info);

  detected_board++;

  printk("fb%d: %s, using %uK of video memory.\n",
	 GET_FB_IDX(l_fb_info->gen.info.node),
	 permedia3_name,
	 (u32)(l_fb_info->fb_size>>10));
}

#ifdef CONFIG_FB_OF
__initfunc(void pm3fb_of_init(struct device_node *dp))
{
  struct pm3fb_info *l_fb_info = &(fb_info[detected_board]);

  DTRACE;

  DPRINTK(1, "OpenFirmware board : %s / %s, %s\n", dp->name, dp->type, dp->full_name);

  {
    long i;
    for (i = 0; i < dp->n_addrs; i++)
      DPRINTK(2, "MemRange : 0x%08x - 0x%x\n", dp->addrs[i].address, dp->addrs[i].size);
  }

  return;

  memset(l_fb_info, 0, sizeof(struct pm3fb_info));
  
  l_fb_info->gen.fbhw = &pm3fb_switch;
  l_fb_info->dev = NULL;
  l_fb_info->dn = dp;

  l_fb_info->pIOBase = (unsigned char*)dp->addrs[3].address;
#ifdef __BIG_ENDIAN
  l_fb_info->pIOBase += PM3_REGS_SIZE;
#endif
  l_fb_info->vIOBase = (unsigned char*)-1;
  l_fb_info->p_fb = (unsigned char*)dp->addrs[1].address;
  l_fb_info->v_fb = (unsigned char*)-1;
  
  l_fb_info->fb_size = (unsigned long) (unsigned char*)dp->addrs[1].size; /* pm3_size_memory(); */

  DPRINTK(2, "OpenFIrmware board : IOBase 0x%08lx, p_fb 0x%08lx, fb_size %d KB\n",
	  (unsigned long)l_fb_info->pIOBase, (unsigned long)l_fb_info->p_fb, l_fb_info->fb_size >> 10);

  pm3_mapIO(l_fb_info);
  
  PM3_SHOW_CUR_MODE
    
  PM3_SHOW_CUR_TIMING

  strcpy(l_fb_info->gen.info.modename, permedia3_name);
  disp.scrollmode=SCROLL_YNOMOVE;
  l_fb_info->gen.parsize=sizeof(struct pm3fb_par);
  l_fb_info->gen.info.changevar = NULL;
  l_fb_info->gen.info.node = -1;
  l_fb_info->gen.info.fbops = &pm3fb_ops;
  l_fb_info->gen.info.disp = &disp;
  l_fb_info->gen.info.switch_con = &fbgen_switch;
  l_fb_info->gen.info.updatevar = &fbgen_update_var; /* */
  l_fb_info->gen.info.blank = &fbgen_blank; /* */
  l_fb_info->gen.info.flags = FBINFO_FLAG_DEFAULT;

  DPRINTK(3, "Var filling done\n");
  
  (void)fbgen_get_var(&disp.var, -1, &l_fb_info->gen.info);
  
  DPRINTK(3, "First fbgen done\n");
  
  (void)fbgen_do_set_var(&disp.var, 1, &l_fb_info->gen);
  
  DPRINTK(3, "Second fbgen done\n");

  fbgen_set_disp(-1, &l_fb_info->gen);

  DPRINTK(3, "Third fbgen done\n");

  fbgen_install_cmap(0, &l_fb_info->gen);

  DPRINTK(3, "Fourth fbgen done\n");

  if (register_framebuffer(&l_fb_info->gen.info) < 0)
	{
	  DPRINTK(1, "Couldn't register framebuffer\n");
	  return;
	}

  DPRINTK(3, "Register fb done\n");
  
  pm3_restore_mode(l_fb_info);

  detected_board++;

  printk("fb%d: %s, using %uK of video memory.\n",
	 GET_FB_IDX(l_fb_info->gen.info.node),
	 permedia3_name,
	 (u32)(l_fb_info->fb_size>>10));
}
#endif

#ifdef MODULE
int init_module(void)
{
  DTRACE;

  pm3fb_init();
  return 0;
}

void cleanup_module(void)
{
  DTRACE;
  {
    unsigned long i;
    pm3fb_info *l_fb_info;
    for (i = 0; i < PM3_MAX_BOARD; i++)
      {
	l_fb_info = &(fb_info[i]);
	if (l_fb_info->vIOBase != (unsigned char*)-1)
	  pm3_unmapIO(l_fb_info);
      }
  }
  unregister_framebuffer(&l_fb_info->gen.info);
  return;
}
#endif /* MODULE */
